import os
import yaml
import click

from subprocess import check_output, CalledProcessError
from jinja2 import Template


class Config():

    @staticmethod
    def env(key):
        """Return the value of the `key` environment variable."""
        try:
            return os.environ[key]
        except KeyError:
            click.secho('[ERROR] No environment variable with key {}'.format(key), err=True, fg='red')
            exit(1)

    @staticmethod
    def git_commit():
        """Return the SHA1 of the latest Git commit (HEAD)."""
        try:
            return check_output(['git', 'rev-parse', '--short', 'HEAD']).strip()
        except CalledProcessError:
            click.secho('[ERROR] Current directory is not a Git repository', err=True, fg='red')
            exit(1)

    @staticmethod
    def git_branch():
        """Return the name of the current Git branch."""
        try:
            return check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip()
        except CalledProcessError:
            click.secho('[ERROR] Current directory is not a Git repository', err=True, fg='red')
            exit(1)

    @staticmethod
    def load(config_file='brume.yml'):
        """
        Return the YAML configuration for a project based on the `config_file` template.

        By default, the template exposes the `env` function.
        The `git_branch` and `git_commit` values are exposed only when a `.git` folder
        exists in the current directory
        """
        template_functions = {}
        template_functions['env'] = Config.env

        if os.path.isdir('.git'):
            template_functions['git_commit'] = Config.git_commit()
            template_functions['git_branch'] = Config.git_branch()

        try:
            template = Template(open(config_file, 'r').read())
        except IOError as e:
            print(e)
            exit(1)
        else:
            return yaml.load(
                template.render(**template_functions)
            )



import time
import os
import boto3
import click

from color import Color
from botocore.exceptions import ClientError

client = boto3.client('cloudformation')


def env(key):
    return os.getenv(key, None)


def make_tags(tags_list):
    return [{"Key": k, "Value": v} for k, v in tags_list.items()]


def make_parameters(tags_list):
    return [{"ParameterKey": k, "ParameterValue": v} for k, v in tags_list.items()]


class Stack():
    stack_name = None
    parameters = {}
    capabilities = []
    tags = {}
    on_failure = 'ROLLBACK'

    def __init__(self, conf):
        self.stack_name = conf['stack_name']
        self.template_body = conf['template_body']
        self.capabilities = conf.get('capabilities', self.capabilities)
        self.parameters = make_parameters(conf.get('parameters', self.parameters))
        self.tags = make_tags(conf['tags'])
        self.on_failure = conf.get('on_failure', self.on_failure)

        self.stack_configuration = dict(
            StackName=self.stack_name,
            TemplateBody=open(self.template_body, 'r').read(),
            Parameters=self.parameters,
            Capabilities=self.capabilities,
            Tags=self.tags)

    def get_stacks(self):
        stacks = [self.stack_name]
        substacks = client.describe_stack_resources(StackName=self.stack_name)['StackResources']
        stacks.extend([s['PhysicalResourceId'] for s in substacks])
        return stacks

    def outputs(self):
        outputs = {}
        try:
            for stack in self.get_stacks():
                s_outputs = client.describe_stacks(StackName=stack)['Stacks'][0].get('Outputs', [])
                for o in s_outputs:
                    outputs[o['OutputKey']] = o['OutputValue']
            return outputs
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            else:
                raise e

    def params(self):
        parameters = {}
        try:
            for stack in self.get_stacks():
                parameters[stack] = {}
                s_params = client.describe_stacks(StackName=stack)['Stacks'][0].get('Parameters', [])
                for o in s_params:
                    parameters[stack][o['ParameterKey']] = o['ParameterValue']
            return parameters
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            else:
                raise e

    @staticmethod
    def exists(stack_name):
        try:
            client.describe_stacks(StackName=stack_name)
        except ClientError as e:
            if 'AlreadyExistsException' in e.message:
                click.secho('Stack [{}] does not exist'.format(stack_name), err=True, fg='red')
                exit(1)
        else:
            return True

    def create(self):
        click.echo('Creating stack {}'.format(self.stack_name))
        try:
            client.create_stack(**self.stack_configuration)
            self.tail()
        except ClientError as e:
            if 'AlreadyExistsException' in e.message:
                click.secho('Stack [{}] already exists'.format(self.stack_name), err=True, fg='red')
                exit(1)

    def update(self):
        click.echo('Updating stack {}'.format(self.stack_name))
        try:
            client.update_stack(**self.stack_configuration)
            self.tail()
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            if 'No updates are to be performed.' in e.message:
                click.secho('No updates are to be performed on stack [{}]'.format(self.stack_name), err=True, fg='red')
                exit(1)

    def create_or_update(self):
        click.echo('Deploying stack {}'.format(self.stack_name))
        try:
            client.create_stack(**self.stack_configuration)
            self.tail()
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            if 'No updates are to be performed.' in e.message:
                click.secho('No updates are to be performed on stack [{}]'.format(self.stack_name), err=True, fg='red')
                exit(1)
            click.secho(e.message, err=True, fg='red')
            click.secho('Stack {} already exists'.format(self.stack_name), err=True, fg='red')
            self.update()

    def delete(self):
        click.echo('Deleting stack {}'.format(self.stack_name))
        try:
            client.delete_stack(StackName=self.stack_name)
            self.tail()
        except ClientError:
            # if 'does not exist' in e.message:
            exit(1)

    def status(self):
        try:
            stacks = client.describe_stacks(StackName=self.stack_name)
            click.echo(Color.for_status(next(s['StackStatus'] for s in stacks['Stacks'])))
        except KeyError as e:
            click.secho(e, err=True, fg='red')
            exit(1)
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)

    def get_events(self):
        Stack.exists(self.stack_name)
        event_list = []
        params = dict(StackName=self.stack_name)
        while True:
            events = client.describe_stack_events(**params)
            event_list.append(events)
            if 'NextToken' not in events:
                break
            else:
                params['NextToken'] = events['NextToken']
                time.sleep(1)
        return reversed(event_list[0]['StackEvents'])

    def tail(self, sleep_time=5):
        click.echo('Polling for events...')
        error = False
        seen = set()
        initial_events = self.get_events()
        self.print_log_headers()
        for e in initial_events:
            self._log_event(e)
            seen.add(e['EventId'])

        while True:
            events = self.get_events()
            for e in events:
                if e['EventId'] not in seen:
                    self._log_event(e)
                    seen.add(e['EventId'])
                    if 'FAILED' in e['ResourceStatus']:
                        error = True
            if self.stack_complete(e):
                if error:
                    exit(1)
                break
            time.sleep(sleep_time)

    def stack_complete(self, e):
        if e['LogicalResourceId'] == self.stack_name and e['ResourceStatus'].endswith('COMPLETE'):
            return True
        else:
            return False

    def print_log_headers(self):
        click.echo('{:23s} {:36s} {:30s} {:30s} {}'.format(
            'Timestamp', 'Status', 'Resource', 'Type', 'Reason'
        ))

    def _log_event(self, e):
        click.echo('{:23s} {:36s} {:30s} {:30s} {}'.format(
            e['Timestamp'].strftime('%Y-%m-%d %H:%M:%S UTC'),
            Color.for_status(e['ResourceStatus']),
            e['LogicalResourceId'],
            e['ResourceType'],
            e.get('ResourceStatusReason', ''),
        ))



import os
import boto3
import sys
import click
from botocore.exceptions import ClientError

s3_client = boto3.client('s3')

CFN_TEMPLATE_SIZE_LIMIT = 51200


class InvalidTemplateError(BaseException):

    def __init__(self, m):
        self.m = m

    def __str__(self):
        return self.m


class Template():
    key = None

    def __init__(self, file, config):
        self.file = file
        self.s3_bucket = config['s3_bucket']
        self.public_url = self.public_url(config)

    def public_url(self, config):
        local_path = config.get('local_path', '')
        if local_path != '.':
            file = self.file.replace(local_path, '')
        else:
            file = self.file
        s3_path = config.get('s3_path', 'cloudformation')
        self.key = os.path.normpath('{}/{}'.format(s3_path, file)).strip('/')
        url = os.path.normpath('{}.s3.amazonaws.com/{}'.format(self.s3_bucket, self.key))
        return 'https://{}'.format(url)

    def content(self):
        try:
            return open(self.file, 'r').read()
        except IOError as e:
            click.secho('File {!r} does not exist'.format(self.file), err=True, fg='red')
            raise e

    def validate(self):
        sys.stdout.write('Validating {} ... '.format(self.file))
        cfn_client = boto3.client('cloudformation')
        if os.path.getsize(self.file) > CFN_TEMPLATE_SIZE_LIMIT:
            params = {'TemplateURL': self.public_url}
        else:
            params = {'TemplateBody': self.content()}
        try:
            cfn_client.validate_template(**params)
        except ClientError as e:
            click.secho('invalid', err=True, fg='red')
            print(e)
            exit(1)
        else:
            click.secho('valid', fg='green')
        return self

    def upload(self):
        print('Publishing {} to {}'.format(self.file, self.public_url))
        s3_client.put_object(
            Bucket=self.s3_bucket,
            Body=self.content(),
            Key=self.key
        )
        return self



import sys
import os
import boto3
import click

s3_client = boto3.client('s3')

def send_assets(local_path, s3_bucket , s3_path= ''):
    """
    Send directory '{local_path}' under 's3://{s3_bucket}/{s3_path}'
    """
    for (dirpath, dirnames, filenames) in os.walk(local_path):
        for filename in filenames:
            sourcepath = os.path.join(dirpath , filename)
            key = s3_path + '/' + os.path.relpath(sourcepath, local_path)
            click.echo('Publishing {} to {}'.format(sourcepath, s3_bucket + '/' + key))
            s3_client.put_object(
                Bucket=s3_bucket,
                Body=open(sourcepath, 'r').read(),
                Key=key
            )



import os
from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='brume',
    version='0.0.10',
    description='AWS Cloudformation deployer',
    long_description=read('README.rst'),
    url='https://github.com/flou/brume',
    author='Lou Ferrand',
    author_email='ferrand@ekino.com',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'boto3==1.4.4',
        'crayons>=0.1.2',
        'click>=6.6',
        'PyYAML>=3.12',
        'Jinja2>=2.8',
    ],
    entry_points={
        'console_scripts': [
            'brume = brume.cli:cli'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



import click

from glob import glob
from os import path
from yaml import dump

from config import Config
from stack import Stack
from template import Template
from assets import send_assets
conf = Config.load('brume.yml')
templates_config = conf['templates']
cf_config = conf['stack']


def process_assets():
    if ('assets' in conf):
        assetsConfig = conf['assets']
        local_path = assetsConfig['local_path']
        s3_bucket = assetsConfig['s3_bucket']
        s3_path = assetsConfig['s3_path']
        click.echo("Processing assets from {} to s3://{}/{}".format(local_path, s3_bucket, s3_path))
        send_assets(local_path, s3_bucket , s3_path)

def collect_templates():
    """Convert every .cform template into a Template."""
    templates = glob(path.join(templates_config.get('local_path', ''), '*.cform'))
    return [Template(t, templates_config) for t in templates]


def validate_and_upload():
    templates = collect_templates()
    map(lambda t: t.validate(), templates)
    map(lambda t: t.upload(), templates)
    process_assets()


@click.command()
def config():
    """Print the current stack confguration."""
    print(dump(conf))


@click.command()
def create():
    """Create a new CloudFormation stack."""
    stack = Stack(cf_config)
    validate_and_upload()
    stack.create()


@click.command()
def update():
    """Update an existing CloudFormation stack."""
    stack = Stack(cf_config)
    validate_and_upload()
    stack.update()


@click.command()
def deploy():
    """Create or update a CloudFormation stack."""
    stack = Stack(cf_config)
    validate_and_upload()
    stack.create_or_update()


@click.command()
def delete():
    """Delete a CloudFormation stack."""
    stack = Stack(cf_config)
    stack.delete()


@click.command()
def status():
    """Get the status of a CloudFormation stack."""
    Stack(cf_config).status()


@click.command()
def outputs():
    """Get the full list of outputs of a CloudFormation stack."""
    outputs = Stack(cf_config).outputs()
    for o in outputs:
        click.echo('{} = {}'.format(o, outputs[o]))


@click.command()
def parameters():
    """Get the full list of parameters of a CloudFormation stack."""
    parameters = Stack(cf_config).params()
    for stack_parameters in parameters:
        sp = parameters[stack_parameters]
        if not sp:
            continue
        click.echo(stack_parameters)
        for p in sp:
            click.echo('\t{} = {}'.format(p, sp[p]))
        click.echo()


@click.command()
def validate():
    """Validate CloudFormation templates."""
    templates = collect_templates()
    return map(lambda t: t.validate(), templates)


@click.command()
def upload():
    """Upload CloudFormation templates and assets to S3."""
    process_assets()
    templates = collect_templates()
    return map(lambda t: t.upload(), templates)


@click.group()
def cli():
    pass


cli.add_command(create)
cli.add_command(update)
cli.add_command(deploy)
cli.add_command(upload)
cli.add_command(delete)
cli.add_command(validate)
cli.add_command(config)
cli.add_command(status)
cli.add_command(outputs)
cli.add_command(parameters)

if __name__ == '__main__':
    cli()
